@@ -11,6 +11,7 @@
 #ifndef RADV_INSTANCE_H
 #define RADV_INSTANCE_H
 
+#include "util/simple_mtx.h"
 #include "util/xmlconfig.h"
 #include "radv_radeon_winsys.h"
 #include "vk_instance.h"
@@ -37,8 +38,11 @@ struct radv_instance {
 
    VkAllocationCallbacks alloc;
 
+   simple_mtx_t shader_dump_mtx;
+
    uint64_t debug_flags;
    uint64_t perftest_flags;
+   uint64_t trap_excp_flags;
    enum radeon_ctx_pstate profile_pstate;
 
    struct {
@@ -50,6 +54,7 @@ struct radv_instance {
       bool disable_shrink_image_store;
       bool disable_aniso_single_level;
       bool disable_trunc_coord;
+      bool disable_depth_storage;
       bool zero_vram;
       bool disable_sinking_load_input_fs;
       bool flush_before_query_copy;
@@ -62,11 +67,11 @@ struct radv_instance {
       bool legacy_sparse_binding;
       bool force_pstate_peak_gfx11_dgpu;
       bool clear_lds;
-      bool enable_dgc;
       bool enable_khr_present_wait;
       bool report_llvm9_version_string;
       bool vk_require_etc2;
       bool vk_require_astc;
+      bool disable_dcc_mips;
       char *app_layer;
       uint8_t override_graphics_shader_version;
       uint8_t override_compute_shader_version;
