@@ -88,9 +88,7 @@ gather_load_fs_input_info(const nir_shader *nir, const nir_intrinsic_instr *intr
    const bool per_primitive = nir->info.per_primitive_inputs & BITFIELD64_BIT(location);
 
    if (!per_primitive) {
-      if (intrin->intrinsic == nir_intrinsic_load_input) {
-         info->ps.flat_shaded_mask |= mapped_mask;
-      } else if (intrin->intrinsic == nir_intrinsic_load_input_vertex) {
+      if (intrin->intrinsic == nir_intrinsic_load_input_vertex) {
          if (io_sem.interp_explicit_strict)
             info->ps.explicit_strict_shaded_mask |= mapped_mask;
          else
@@ -100,6 +98,8 @@ gather_load_fs_input_info(const nir_shader *nir, const nir_intrinsic_instr *intr
             info->ps.float16_hi_shaded_mask |= mapped_mask;
          else
             info->ps.float16_shaded_mask |= mapped_mask;
+      } else if (intrin->intrinsic == nir_intrinsic_load_interpolated_input) {
+         info->ps.float32_shaded_mask |= mapped_mask;
       }
    }
 
@@ -270,6 +270,9 @@ gather_intrinsic_info(const nir_shader *nir, const nir_intrinsic_instr *instr, s
       }
       break;
    }
+   case nir_intrinsic_load_pixel_coord:
+      info->ps.reads_pixel_coord = true;
+      break;
    case nir_intrinsic_load_frag_coord:
       info->ps.reads_frag_coord_mask |= nir_def_components_read(&instr->def);
       break;
@@ -294,6 +297,7 @@ gather_intrinsic_info(const nir_shader *nir, const nir_intrinsic_instr *instr, s
       break;
    }
    case nir_intrinsic_load_input:
+   case nir_intrinsic_load_per_primitive_input:
    case nir_intrinsic_load_interpolated_input:
    case nir_intrinsic_load_input_vertex:
       gather_intrinsic_load_input_info(nir, instr, info, gfx_state, stage_key);
@@ -393,6 +397,69 @@ assign_outinfo_params(struct radv_vs_output_info *outinfo, uint64_t mask, unsign
    }
 }
 
+static void
+radv_get_output_masks(const struct nir_shader *nir, const struct radv_graphics_state_key *gfx_state,
+                      uint64_t *per_vtx_mask, uint64_t *per_prim_mask)
+{
+   /* These are not compiled into neither output param nor position exports. */
+   const uint64_t special_mask = BITFIELD64_BIT(VARYING_SLOT_PRIMITIVE_COUNT) |
+                                 BITFIELD64_BIT(VARYING_SLOT_PRIMITIVE_INDICES) |
+                                 BITFIELD64_BIT(VARYING_SLOT_CULL_PRIMITIVE);
+
+   *per_prim_mask = nir->info.outputs_written & nir->info.per_primitive_outputs & ~special_mask;
+   *per_vtx_mask = nir->info.outputs_written & ~nir->info.per_primitive_outputs & ~special_mask;
+
+   /* Mesh multiview is only lowered in ac_nir_lower_ngg, so we have to fake it here. */
+   if (nir->info.stage == MESA_SHADER_MESH && gfx_state->has_multiview_view_index)
+      *per_prim_mask |= VARYING_BIT_LAYER;
+}
+
+static void
+radv_set_vs_output_param(struct radv_device *device, const struct nir_shader *nir,
+                         const struct radv_graphics_state_key *gfx_state, struct radv_shader_info *info,
+                         bool export_prim_id, bool export_clip_cull_dists)
+{
+   const struct radv_physical_device *pdev = radv_device_physical(device);
+   struct radv_vs_output_info *outinfo = &info->outinfo;
+   uint64_t per_vtx_mask, per_prim_mask;
+
+   radv_get_output_masks(nir, gfx_state, &per_vtx_mask, &per_prim_mask);
+
+   memset(outinfo->vs_output_param_offset, AC_EXP_PARAM_UNDEFINED, sizeof(outinfo->vs_output_param_offset));
+
+   unsigned total_param_exports = 0;
+
+   /* Per-vertex outputs */
+   assign_outinfo_params(outinfo, per_vtx_mask, &total_param_exports, 0);
+
+   if (export_prim_id && (nir->info.stage == MESA_SHADER_VERTEX || nir->info.stage == MESA_SHADER_TESS_EVAL)) {
+      /* Mark the primitive ID as output when it's implicitly exported by VS or TES. */
+      if (outinfo->vs_output_param_offset[VARYING_SLOT_PRIMITIVE_ID] == AC_EXP_PARAM_UNDEFINED)
+         outinfo->vs_output_param_offset[VARYING_SLOT_PRIMITIVE_ID] = total_param_exports++;
+
+      outinfo->export_prim_id = true;
+   }
+
+   if (export_clip_cull_dists) {
+      if (nir->info.outputs_written & VARYING_BIT_CLIP_DIST0)
+         outinfo->vs_output_param_offset[VARYING_SLOT_CLIP_DIST0] = total_param_exports++;
+      if (nir->info.outputs_written & VARYING_BIT_CLIP_DIST1)
+         outinfo->vs_output_param_offset[VARYING_SLOT_CLIP_DIST1] = total_param_exports++;
+   }
+
+   outinfo->param_exports = total_param_exports;
+
+   /* The HW always assumes that there is at least 1 per-vertex param.
+    * so if there aren't any, we have to offset per-primitive params by 1.
+    */
+   const unsigned extra_offset = !!(total_param_exports == 0 && pdev->info.gfx_level >= GFX11);
+
+   /* Per-primitive outputs: the HW needs these to be last. */
+   assign_outinfo_params(outinfo, per_prim_mask, &total_param_exports, extra_offset);
+
+   outinfo->prim_param_exports = total_param_exports - outinfo->param_exports;
+}
+
 static uint8_t
 radv_get_wave_size(struct radv_device *device, gl_shader_stage stage, const struct radv_shader_info *info,
                    const struct radv_shader_stage_key *stage_key)
@@ -444,6 +511,9 @@ radv_compute_esgs_itemsize(const struct radv_device *device, uint32_t num_varyin
 static void
 gather_shader_info_ngg_query(struct radv_device *device, struct radv_shader_info *info)
 {
+   const struct radv_physical_device *pdev = radv_device_physical(device);
+
+   info->gs.has_pipeline_stat_query = pdev->emulate_ngg_gs_query_pipeline_stat && info->stage == MESA_SHADER_GEOMETRY;
    info->has_xfb_query = info->so.num_outputs > 0;
    info->has_prim_query = device->cache_key.primitives_generated_query || info->has_xfb_query;
 }
@@ -492,6 +562,9 @@ gather_shader_info_vs(struct radv_device *device, const nir_shader *nir,
       info->vs.dynamic_inputs = true;
    }
 
+   info->gs_inputs_read = ~0ULL;
+   info->vs.hs_inputs_read = ~0ULL;
+
    /* Use per-attribute vertex descriptors to prevent faults and for correct bounds checking. */
    info->vs.use_per_attribute_vb_descs = radv_use_per_attribute_vb_descs(nir, gfx_state, stage_key);
 
@@ -573,6 +646,7 @@ gather_shader_info_tcs(struct radv_device *device, const nir_shader *nir,
 static void
 gather_shader_info_tes(struct radv_device *device, const nir_shader *nir, struct radv_shader_info *info)
 {
+   info->gs_inputs_read = ~0ULL;
    info->tes._primitive_mode = nir->info.tess._primitive_mode;
    info->tes.spacing = nir->info.tess.spacing;
    info->tes.ccw = nir->info.tess.ccw;
@@ -740,7 +814,6 @@ radv_get_legacy_gs_info(const struct radv_device *device, struct radv_shader_inf
 static void
 gather_shader_info_gs(struct radv_device *device, const nir_shader *nir, struct radv_shader_info *info)
 {
-   const struct radv_physical_device *pdev = radv_device_physical(device);
    unsigned add_clip = nir->info.clip_distance_array_size + nir->info.cull_distance_array_size > 4;
    info->gs.gsvs_vertex_size = (util_bitcount64(nir->info.outputs_written) + add_clip) * 16;
    info->gs.max_gsvs_emit_size = info->gs.gsvs_vertex_size * nir->info.gs.vertices_out;
@@ -751,7 +824,6 @@ gather_shader_info_gs(struct radv_device *device, const nir_shader *nir, struct
    info->gs.output_prim = nir->info.gs.output_primitive;
    info->gs.invocations = nir->info.gs.invocations;
    info->gs.max_stream = nir->info.gs.active_stream_mask ? util_last_bit(nir->info.gs.active_stream_mask) - 1 : 0;
-   info->gs.has_pipeline_stat_query = pdev->emulate_ngg_gs_query_pipeline_stat;
 
    for (unsigned slot = 0; slot < VARYING_SLOT_MAX; ++slot) {
       const uint8_t usage_mask = info->gs.output_usage_mask[slot];
@@ -870,6 +942,7 @@ gather_shader_info_fs(const struct radv_device *device, const nir_shader *nir,
    info->ps.post_depth_coverage = nir->info.fs.post_depth_coverage;
    info->ps.depth_layout = nir->info.fs.depth_layout;
    info->ps.uses_sample_shading = nir->info.fs.uses_sample_shading;
+   info->ps.uses_fbfetch_output = nir->info.fs.uses_fbfetch_output;
    info->ps.writes_memory = nir->info.writes_memory;
    info->ps.has_pcoord = nir->info.inputs_read & VARYING_BIT_PNTC;
    info->ps.prim_id_input = nir->info.inputs_read & VARYING_BIT_PRIMITIVE_ID;
@@ -893,6 +966,7 @@ gather_shader_info_fs(const struct radv_device *device, const nir_shader *nir,
       !(uses_persp_or_linear_interp || info->ps.needs_sample_positions || info->ps.reads_frag_shading_rate ||
         info->ps.writes_memory || nir->info.fs.needs_quad_helper_invocations ||
         BITSET_TEST(nir->info.system_values_read, SYSTEM_VALUE_FRAG_COORD) ||
+        BITSET_TEST(nir->info.system_values_read, SYSTEM_VALUE_PIXEL_COORD) ||
         BITSET_TEST(nir->info.system_values_read, SYSTEM_VALUE_POINT_COORD) ||
         BITSET_TEST(nir->info.system_values_read, SYSTEM_VALUE_SAMPLE_ID) ||
         BITSET_TEST(nir->info.system_values_read, SYSTEM_VALUE_SAMPLE_POS) ||
@@ -909,9 +983,9 @@ gather_shader_info_fs(const struct radv_device *device, const nir_shader *nir,
       info->ps.spi_ps_input_addr &= C_02865C_COVERAGE_TO_SHADER_SELECT;
    }
 
-   info->has_epilog = gfx_state->ps.has_epilog && info->ps.colors_written;
+   info->ps.has_epilog = gfx_state->ps.has_epilog && info->ps.colors_written;
 
-   if (!info->has_epilog) {
+   if (!info->ps.has_epilog) {
       info->ps.mrt0_is_dual_src = gfx_state->ps.epilog.mrt0_is_dual_src;
       info->ps.spi_shader_col_format = gfx_state->ps.epilog.spi_shader_col_format;
 
@@ -925,7 +999,7 @@ gather_shader_info_fs(const struct radv_device *device, const nir_shader *nir,
       (info->ps.color0_written & 0x8) && (info->ps.writes_z || info->ps.writes_stencil || info->ps.writes_sample_mask);
 
    info->ps.exports_mrtz_via_epilog =
-      info->has_epilog && gfx_state->ps.exports_mrtz_via_epilog && export_alpha_and_mrtz;
+      info->ps.has_epilog && gfx_state->ps.exports_mrtz_via_epilog && export_alpha_and_mrtz;
 
    if (!info->ps.exports_mrtz_via_epilog) {
       info->ps.writes_mrt0_alpha = gfx_state->ms.alpha_to_coverage_via_mrtz && export_alpha_and_mrtz;
@@ -1142,19 +1216,13 @@ radv_nir_shader_info_pass(struct radv_device *device, const struct nir_shader *n
    if (nir->info.stage == MESA_SHADER_VERTEX || nir->info.stage == MESA_SHADER_TESS_EVAL ||
        nir->info.stage == MESA_SHADER_GEOMETRY || nir->info.stage == MESA_SHADER_MESH) {
       struct radv_vs_output_info *outinfo = &info->outinfo;
+      uint64_t per_vtx_mask, per_prim_mask;
 
-      /* These are not compiled into neither output param nor position exports. */
-      uint64_t special_mask = BITFIELD64_BIT(VARYING_SLOT_PRIMITIVE_COUNT) |
-                              BITFIELD64_BIT(VARYING_SLOT_PRIMITIVE_INDICES) |
-                              BITFIELD64_BIT(VARYING_SLOT_CULL_PRIMITIVE);
-      uint64_t per_prim_mask = nir->info.outputs_written & nir->info.per_primitive_outputs & ~special_mask;
-      uint64_t per_vtx_mask = nir->info.outputs_written & ~nir->info.per_primitive_outputs & ~special_mask;
+      radv_get_output_masks(nir, gfx_state, &per_vtx_mask, &per_prim_mask);
 
-      /* Mesh multivew is only lowered in ac_nir_lower_ngg, so we have to fake it here. */
-      if (nir->info.stage == MESA_SHADER_MESH && gfx_state->has_multiview_view_index) {
-         per_prim_mask |= VARYING_BIT_LAYER;
+      /* Mesh multiview is only lowered in ac_nir_lower_ngg, so we have to fake it here. */
+      if (nir->info.stage == MESA_SHADER_MESH && gfx_state->has_multiview_view_index)
          info->uses_view_index = true;
-      }
 
       /* Per vertex outputs. */
       outinfo->writes_pointsize = per_vtx_mask & VARYING_BIT_PSIZ;
@@ -1188,25 +1256,6 @@ radv_nir_shader_info_pass(struct radv_device *device, const struct nir_shader *n
          pos_written |= 1 << 3;
 
       outinfo->pos_exports = util_bitcount(pos_written);
-
-      memset(outinfo->vs_output_param_offset, AC_EXP_PARAM_UNDEFINED, sizeof(outinfo->vs_output_param_offset));
-
-      unsigned total_param_exports = 0;
-
-      /* Per-vertex outputs */
-      assign_outinfo_params(outinfo, per_vtx_mask, &total_param_exports, 0);
-
-      outinfo->param_exports = total_param_exports;
-
-      /* The HW always assumes that there is at least 1 per-vertex param.
-       * so if there aren't any, we have to offset per-primitive params by 1.
-       */
-      const unsigned extra_offset = !!(total_param_exports == 0 && pdev->info.gfx_level >= GFX11);
-
-      /* Per-primitive outputs: the HW needs these to be last. */
-      assign_outinfo_params(outinfo, per_prim_mask, &total_param_exports, extra_offset);
-
-      outinfo->prim_param_exports = total_param_exports - outinfo->param_exports;
    }
 
    info->vs.needs_draw_id |= BITSET_TEST(nir->info.system_values_read, SYSTEM_VALUE_DRAW_ID);
@@ -1647,12 +1696,22 @@ radv_determine_ngg_settings(struct radv_device *device, struct radv_shader_stage
                             struct radv_shader_stage *fs_stage, const struct radv_graphics_state_key *gfx_state)
 {
    const struct radv_physical_device *pdev = radv_device_physical(device);
+   uint64_t ps_inputs_read;
 
    assert(es_stage->stage == MESA_SHADER_VERTEX || es_stage->stage == MESA_SHADER_TESS_EVAL);
    assert(!fs_stage || fs_stage->stage == MESA_SHADER_FRAGMENT);
 
-   /* NGG culling is implicitly disabled when the FS stage is unknown. */
-   uint64_t ps_inputs_read = fs_stage ? fs_stage->nir->info.inputs_read : ~0;
+   if (fs_stage) {
+      ps_inputs_read = fs_stage->nir->info.inputs_read;
+   } else {
+      /* Rely on the number of VS/TES outputs when the FS is unknown (for fast-link or unlinked ESO)
+       * because this should be a good approximation of the number of FS inputs.
+       */
+      ps_inputs_read = es_stage->nir->info.outputs_written;
+
+      /* Clear varyings that can't be PS inputs. */
+      ps_inputs_read &= ~(VARYING_BIT_POS | VARYING_BIT_PSIZ);
+   }
 
    unsigned num_vertices_per_prim = 0;
    if (es_stage->stage == MESA_SHADER_VERTEX) {
@@ -1687,24 +1746,10 @@ radv_link_shaders_info(struct radv_device *device, struct radv_shader_stage *pro
     */
    if (producer->info.next_stage == MESA_SHADER_FRAGMENT ||
        !(gfx_state->lib_flags & VK_GRAPHICS_PIPELINE_LIBRARY_FRAGMENT_SHADER_BIT_EXT)) {
-      struct radv_vs_output_info *outinfo = &producer->info.outinfo;
       const bool ps_prim_id_in = !consumer || consumer->info.ps.prim_id_input;
       const bool ps_clip_dists_in = !consumer || !!consumer->info.ps.input_clips_culls_mask;
 
-      if (ps_prim_id_in && (producer->stage == MESA_SHADER_VERTEX || producer->stage == MESA_SHADER_TESS_EVAL)) {
-         /* Mark the primitive ID as output when it's implicitly exported by VS or TES. */
-         if (outinfo->vs_output_param_offset[VARYING_SLOT_PRIMITIVE_ID] == AC_EXP_PARAM_UNDEFINED)
-            outinfo->vs_output_param_offset[VARYING_SLOT_PRIMITIVE_ID] = outinfo->param_exports++;
-
-         outinfo->export_prim_id = true;
-      }
-
-      if (ps_clip_dists_in) {
-         if (producer->nir->info.outputs_written & VARYING_BIT_CLIP_DIST0)
-            outinfo->vs_output_param_offset[VARYING_SLOT_CLIP_DIST0] = outinfo->param_exports++;
-         if (producer->nir->info.outputs_written & VARYING_BIT_CLIP_DIST1)
-            outinfo->vs_output_param_offset[VARYING_SLOT_CLIP_DIST1] = outinfo->param_exports++;
-      }
+      radv_set_vs_output_param(device, producer->nir, gfx_state, &producer->info, ps_prim_id_in, ps_clip_dists_in);
    }
 
    if (producer->stage == MESA_SHADER_VERTEX || producer->stage == MESA_SHADER_TESS_EVAL) {
@@ -1725,12 +1770,18 @@ radv_link_shaders_info(struct radv_device *device, struct radv_shader_stage *pro
 
          es_info->workgroup_size = gs_info->workgroup_size;
       }
+
+      if (consumer && consumer->stage == MESA_SHADER_GEOMETRY) {
+         producer->info.gs_inputs_read = consumer->nir->info.inputs_read;
+      }
    }
 
    if (producer->stage == MESA_SHADER_VERTEX && consumer && consumer->stage == MESA_SHADER_TESS_CTRL) {
       struct radv_shader_stage *vs_stage = producer;
       struct radv_shader_stage *tcs_stage = consumer;
 
+      vs_stage->info.vs.hs_inputs_read = tcs_stage->nir->info.inputs_read;
+
       if (gfx_state->ts.patch_control_points) {
          vs_stage->info.workgroup_size =
             ac_compute_lshs_workgroup_size(pdev->info.gfx_level, MESA_SHADER_VERTEX, tcs_stage->info.num_tess_patches,
